import sys

LOG_LEVEL = "INFO"
LOG_FORMAT_HEADER = "[%(asctime)s][%(levelname)7s]"
LOG_FORMAT = LOG_FORMAT_HEADER + " %(message)s"
LOGGING_CONFIG = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "shell": {"format": LOG_FORMAT},
    },
    "handlers": {
        "console": {
            "class": "airflow_db_logger.handlers.StreamHandler",
            "formatter": "shell",
        },
        "task": {
            "class": "airflow_db_logger.handlers.StreamHandler",
            "formatter": "shell",
        },
        "processor": {
            "class": "airflow_db_logger.handlers.StreamHandler",
            "formatter": "shell",
            "level": "WARN",
        },
    },
    "loggers": {
        "airflow.processor": {
            "handlers": ["processor"],
            "level": LOG_LEVEL,
            "propagate": False,
        },
        "airflow.task": {
            "handlers": ["task"],
            "level": LOG_LEVEL,
            "propagate": False,
        },
        "flask_appbuilder": {
            "handler": ["console"],
            "level": LOG_LEVEL,
            "propagate": True,
        },
    },
    "root": {
        "handlers": ["console"],
        "level": LOG_LEVEL,
    },
}



from copy import deepcopy
from airflow_db_logger.shell_logging_config import LOGGING_CONFIG

global IS_LOADING_CONFIG
IS_LOADING_CONFIG = False

global DB_LOGGER_LOGGING_CONFIG
DB_LOGGER_LOGGING_CONFIG = deepcopy(LOGGING_CONFIG)



import logging
import airflow
from datetime import datetime
from airflow import settings
from airflow.operators.python_operator import PythonOperator
from sqlalchemy.orm import Session, Query

from airflow_db_logger.config import DBLoggerSession
from airflow_db_logger.data import TaskExecutionLogRecord, DagFileProcessingLogRecord, LoggerModelBase


def create_clean_old_logs_task(
    task_id: str,
    before: datetime,
    after: datetime = None,
    dag: airflow.DAG = None,
):
    dag: airflow.DAG = dag or settings.CONTEXT_MANAGER_DAG
    assert isinstance(dag, airflow.DAG), ValueError("dag is none and not using 'with dag as ..' pattern")
    assert task_id and isinstance(task_id, str), ValueError("Task id must be a non empty string")

    after = after or datetime(year=2000, month=1, day=1)

    def create_select_query(session: Session, cls: LoggerModelBase) -> Query:
        query: Query = session.query(cls)
        query = query.filter(cls.timestamp <= before)
        query = query.filter(cls.timestamp > after)
        return query

    def clean_airflow_logs(*args, **kwargs):
        logging.info(f"Cleaning up logs from {after} to {before}")
        session: Session = DBLoggerSession()
        task_logs = create_select_query(session, TaskExecutionLogRecord)
        processor_logs = create_select_query(session, DagFileProcessingLogRecord)

        task_logs.delete()
        processor_logs.delete()

        session.commit()

    return PythonOperator(
        task_id=task_id,
        python_callable=clean_airflow_logs,
    )



import os
import logging
from typing import Dict
from weakref import WeakValueDictionary
from zthreading.decorators import collect_delayed_calls_async
from airflow_db_logger.exceptions import DBLoggerException
from airflow_db_logger.config import BASE_LOG_FOLDER
from airflow_db_logger.handlers import DBLogStreamWriter, DBLogHandler, stderr_logger


class DBLogFileCollectionWriter:
    def __init__(self, filename: str) -> None:
        self.filename = filename
        self.pending_records = []

    def write(self, record):
        self.pending_records.append(record)
        self._write_async()

    @collect_delayed_calls_async(interval=0.05, on_error="write_async_error")
    def _write_async(self):
        if len(self.pending_records) == 0:
            return
        try:
            records = self.pending_records
            self.pending_records = []
            records = [str(r) for r in records]
            dirname = os.path.dirname(self.filename)
            if not os.path.isdir(dirname):
                os.makedirs(dirname)
            with open(self.filename, "a") as logfile:
                logfile.write("\n".join(records))
                logfile.write("\n")

        except Exception as err:
            stderr_logger.error("Failed to flash to file @ " + self.filename)
            stderr_logger.error(err)

    def write_async_error(self, err: Exception):
        stderr_logger.error(err)


class DBLogFileWriter(DBLogStreamWriter):
    def __init__(self) -> None:
        super().__init__()
        self._pending_loggers: Dict[str, DBLogFileCollectionWriter] = WeakValueDictionary()

    def get_file_collection_writer(self, filename: str):

        if filename not in self._pending_loggers:
            logger = DBLogFileCollectionWriter(filename)
            self._pending_loggers[filename] = logger
            return logger
        else:
            return self._pending_loggers[filename]

    def write(self, handler: DBLogHandler, record: logging.LogRecord):
        # Only applies when using context execution.
        if not handler.has_context:
            return

        logfile = handler.get_logfile_subpath()
        assert isinstance(logfile, str), DBLoggerException(
            f"Invalid logfile when writing log from {type(handler)}: {logfile}"
        )
        filename = os.path.join(BASE_LOG_FOLDER, handler.get_logfile_subpath())
        self.get_file_collection_writer(filename=filename).write(handler.format(record))



from sqlalchemy import Column, Integer, String, Text, Index, DateTime, MetaData
from sqlalchemy.ext.declarative import declarative_base
from datetime import datetime
from airflow.utils.sqlalchemy import UtcDateTime
from airflow_db_logger.config import DB_LOGGER_SQL_ALCHEMY_SCHEMA, DB_LOGGER_CREATE_INDEXES

LoggerModelBase = declarative_base(
    metadata=(
        None
        if not DB_LOGGER_SQL_ALCHEMY_SCHEMA or DB_LOGGER_SQL_ALCHEMY_SCHEMA.isspace()
        else MetaData(schema=DB_LOGGER_SQL_ALCHEMY_SCHEMA)
    )
)


def create_table_indexes(*args):
    if not DB_LOGGER_CREATE_INDEXES:
        return []
    return args


class TaskExecutionLogRecord(LoggerModelBase):
    # FIXME: Not the base name for this table,
    # but since log was taken, it will have to do.
    # NOTE: This class is very similar to airflow.models.log,
    # but differs in purpose. Since we want
    # indexing to allow for fast log retrival, airflow.models.log
    # was not used.
    __tablename__ = "log_task_execution"

    id = Column(Integer, primary_key=True)
    timestamp = Column(DateTime)
    dag_id = Column(String)
    task_id = Column(String)
    execution_date = Column(UtcDateTime)
    try_number = Column(Integer)
    text = Column(Text)

    def __init__(
        self,
        dag_id: str,
        task_id: str,
        execution_date: datetime,
        try_number: int,
        text: str,
    ):
        super().__init__()
        self.timestamp = datetime.now()

        self.dag_id = dag_id
        self.task_id = task_id
        self.execution_date = execution_date
        self.try_number = try_number
        self.text = text


class DagFileProcessingLogRecord(LoggerModelBase):
    # NOTE: This class is very similar to airflow.models.log,
    # but differs in purpose. Since we want
    # indexing to allow for fast log retrival, airflow.models.log
    # was not used.
    __tablename__ = "log_dag_processing"

    id = Column(Integer, primary_key=True)
    timestamp = Column(DateTime)
    dag_filename = Column(String)
    text = Column(Text)

    def __init__(
        self,
        dag_filename: str,
        text: str,
    ):
        super().__init__()

        self.timestamp = datetime.now()

        self.dag_filename = dag_filename
        self.text = text


if DB_LOGGER_CREATE_INDEXES:
    Index("task_execution_log_record_timestamp_idx", TaskExecutionLogRecord.timestamp)
    Index("task_execution_log_record_dag_id_idx", TaskExecutionLogRecord.dag_id)
    Index("task_execution_log_record_task_id_idx", TaskExecutionLogRecord.task_id)
    Index("task_execution_log_record_execution_date_idx", TaskExecutionLogRecord.execution_date)
    Index("task_execution_log_record_try_number_idx", TaskExecutionLogRecord.try_number)

    Index("dag_file_processing_log_record_timestamp_idx", DagFileProcessingLogRecord.timestamp)
    Index("dag_file_processing_log_record_dag_filename_idx", DagFileProcessingLogRecord.dag_filename)



import os
import sys
import inspect
from typing import Type

global IS_LOADING_CONFIG
IS_LOADING_CONFIG = False


def get_is_no_color():
    val = os.environ.get("NO_COLOR", "--no-color" in sys.argv)
    if not isinstance(val, bool):
        val = val.strip().lower()
        os.environ["NO_COLOR"] = val
    return val


def colorize(val, color, add_reset=True):
    if get_is_no_color():
        return val
    return color + val + ("\033[0m" if add_reset else "")


def get_calling_frame_objects_by_type(otype: Type, offset: int = 1, first_only=False):
    frames = inspect.stack()
    frames = frames[offset:]
    lst = []
    for f in frames:
        if "self" not in f.frame.f_locals:
            continue
        obj = f.frame.f_locals["self"]
        if not isinstance(obj, otype):
            continue
        if first_only:
            return obj
        lst.append(obj)

    if first_only:
        return None

    return lst


class style:
    GRAY = lambda x: colorize(str(x), "\033[90m")  # noqa: E731
    LIGHT_GRAY = lambda x: colorize(str(x), "\033[37m")  # noqa: E731
    BLACK = lambda x: colorize(str(x), "\033[30m")  # noqa: E731
    RED = lambda x: colorize(str(x), "\033[31m")  # noqa: E731
    GREEN = lambda x: colorize(str(x), "\033[32m")  # noqa: E731
    YELLOW = lambda x: colorize(str(x), "\033[33m")  # noqa: E731
    BLUE = lambda x: colorize(str(x), "\033[34m")  # noqa: E731
    MAGENTA = lambda x: colorize(str(x), "\033[35m")  # noqa: E731
    CYAN = lambda x: colorize(str(x), "\033[36m")  # noqa: E731
    WHITE = lambda x: colorize(str(x), "\033[97m")  # noqa: E731
    UNDERLINE = lambda x: colorize(str(x), "\033[4m")  # noqa: E731
    RESET = lambda x: colorize(str(x), "\033[0m")  # noqa: E731



import sys
import os
import logging
import warnings
from typing import Type
from enum import Enum
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker, scoped_session
from sqlalchemy.pool import NullPool, QueuePool
from airflow.configuration import conf, AirflowConfigException, log

AIRFLOW_CONFIG_SECTION_NAME = "db_logger"


def conf_get_no_warnings_no_errors(*args, **kwargs):
    old_level = log.level
    log.level = logging.ERROR
    try:
        val = conf.get(*args, **kwargs)
    except AirflowConfigException:
        val = None
    log.level = old_level
    return val


def get(
    key: str,
    default=None,
    otype: Type = None,
    allow_empty: bool = False,
    collection: str = None,
):
    otype = otype or str if default is None else default.__class__
    collection = collection or AIRFLOW_CONFIG_SECTION_NAME
    val = conf_get_no_warnings_no_errors(collection, key)

    if issubclass(otype, Enum):
        allow_empty = False

    value_is_empty = val is None or (isinstance(val, str) and len(val.strip()) == 0)

    if not allow_empty and value_is_empty:
        assert default is not None, f"Airflow configuration {collection}.{key} not found, and no default value"
        return default

    if val is None:
        return None
    if otype == bool:
        return val.lower() == "true"
    elif issubclass(otype, Enum):
        val = val.strip()
        return otype(val.strip())
    else:
        return otype(val)


# Loading airflow parameters
LOG_LEVEL = get(collection="core", key="logging_level").upper()
FILENAME_TEMPLATE = get(collection="core", key="LOG_FILENAME_TEMPLATE")
AIRFLOW_EXECUTOR = get(collection="core", key="executor")
IS_RUNNING_DEBUG_EXECUTOR = AIRFLOW_EXECUTOR == "DebugExecutor"
IS_USING_COLORED_CONSOLE = get(collection="core", key="colored_console_log").lower() == "true"

# Loading sql parameters
SQL_ALCHEMY_CONN = get(collection="core", key="sql_alchemy_conn", allow_empty=False)
DAGS_FOLDER = os.path.expanduser(get(collection="core", key="dags_folder"))
BASE_LOG_FOLDER = os.path.expanduser(get(collection="core", key="base_log_folder"))
SQL_ALCHEMY_SCHEMA = get(collection="core", key="sql_alchemy_schema", allow_empty=True)
COLORED_CONSOLE_LOG = get(collection="core", key="colored_console_log")
TASK_LOG_FILENAME_TEMPLATE = (
    get(
        collection="core",
        key="LOG_FILENAME_TEMPLATE",
        default="{{ ti.dag_id }}/{{ ti.task_id }}/{{ ts }}/{{ try_number }}.log",
    )
    .replace("{{{{", "{{")
    .replace("}}}}", "}}")
)
PROCESS_LOG_FILENAME_TEMPLATE = (
    get(
        collection="core",
        key="log_processor_filename_template",
        default="{{ filename }}.log",
    )
    .replace("{{{{", "{{")
    .replace("}}}}", "}}")
)


DB_LOGGER_SQL_ALCHEMY_SCHEMA = get("sql_alchemy_schema", SQL_ALCHEMY_SCHEMA)
DB_LOGGER_SQL_ALCHEMY_CONNECTION = get("sql_alchemy_conn", SQL_ALCHEMY_CONN)
DB_LOGGER_SHOW_REVERSE_ORDER = get("show_reverse", False)
DB_LOGGER_SQL_ALCHEMY_CONNECTION_ARGS = get("sql_alchemy_conn_args", None, allow_empty=True)
DB_LOGGER_CREATE_INDEXES = get("create_index", True)
DB_LOGGER_WRITE_DAG_PROCESSING_TO_DB = get("write_dag_processing_to_db", False)

DB_LOGGER_SQL_ALCHEMY_POOL_ENABLED = get("sql_alchemy_pool_enabled", False)
DB_LOGGER_SQL_ALCHEMY_POOL_SIZE = get("sql_alchemy_pool_size", 5)
DB_LOGGER_SQL_ALCHEMY_MAX_OVERFLOW = get("sql_alchemy_max_overflow", 1)
DB_LOGGER_SQL_ALCHEMY_POOL_RECYCLE = get("sql_alchemy_pool_recycle", 1800)
DB_LOGGER_SQL_ALCHEMY_POOL_PRE_PING = get("sql_alchemy_pool_pre_ping", True)
DB_LOGGER_SQL_ENGINE_ENCODING = get("sql_engine_encoding", "utf-8")

DB_LOGGER_GOOGLE_APP_CREDS_PATH = get("google_application_credentials", default=None, allow_empty=True, otype=str)
# A bucket path, requires google-cloud-storage to be installed.
DB_LOGGER_WRITE_TO_GCS_BUCKET = get("write_to_gcs_bucket", default=None, allow_empty=True, otype=str)
DB_LOGGER_WRITE_TO_GCS_PROJECT_ID = get("write_to_gcs_project_id", default=None, allow_empty=True, otype=str)
DB_LOGGER_WRITE_TO_GCS_MULTI_FILE_LOG = get("write_to_gcs_multi_file_log", default=False)
DB_LOGGER_PROCESSER_LOG_LEVEL = get("processer_log_level", default="WARN", allow_empty=True, otype=str)
# True or path
DB_LOGGER_WRITE_TO_FILES = get("write_to_files", default=False)
# True or path
DB_LOGGER_WRITE_TO_SHELL = get("write_to_shell", default=IS_RUNNING_DEBUG_EXECUTOR)

DB_LOGGER_CONSOLE_FORMATTER = "airflow_coloured" if IS_USING_COLORED_CONSOLE else "airflow"
DB_LOGGER_TASK_FORMATTER = "airflow_coloured" if IS_RUNNING_DEBUG_EXECUTOR and IS_USING_COLORED_CONSOLE else "airflow"


# Setting the default logger log level
logging.basicConfig(level=LOG_LEVEL)

logging.debug(f"DBLogger is connecting to: {DB_LOGGER_SQL_ALCHEMY_CONNECTION}/{DB_LOGGER_SQL_ALCHEMY_SCHEMA}")
logging.debug(f"DBLogger indexes: {DB_LOGGER_CREATE_INDEXES}")


def create_db_logger_sqlalchemy_engine():

    # Configuring the db_logger sql engine.
    engine_args = {}
    if DB_LOGGER_SQL_ALCHEMY_POOL_ENABLED:
        # Copied from airflow main repo.

        # Pool size engine args not supported by sqlite.
        # If no config value is defined for the pool size, select a reasonable value.
        # 0 means no limit, which could lead to exceeding the Database connection limit.
        pool_size = DB_LOGGER_SQL_ALCHEMY_POOL_SIZE

        # The maximum overflow size of the pool.
        # When the number of checked-out connections reaches the size set in pool_size,
        # additional connections will be returned up to this limit.
        # When those additional connections are returned to the pool, they are disconnected and discarded.
        # It follows then that the total number of simultaneous connections
        # the pool will allow is pool_size + max_overflow,
        # and the total number of “sleeping” connections the pool will allow is pool_size.
        # max_overflow can be set to -1 to indicate no overflow limit;
        # no limit will be placed on the total number
        # of concurrent connections. Defaults to 10.
        max_overflow = DB_LOGGER_SQL_ALCHEMY_MAX_OVERFLOW

        # The DB server already has a value for wait_timeout (number of seconds after
        # which an idle sleeping connection should be killed). Since other DBs may
        # co-exist on the same server, SQLAlchemy should set its
        # pool_recycle to an equal or smaller value.
        pool_recycle = DB_LOGGER_SQL_ALCHEMY_POOL_RECYCLE

        # Check connection at the start of each connection pool checkout.
        # Typically, this is a simple statement like “SELECT 1”, but may also make use
        # of some DBAPI-specific method to test the connection for liveness.
        # More information here:
        # https://docs.sqlalchemy.org/en/13/core/pooling.html#disconnect-handling-pessimistic
        pool_pre_ping = DB_LOGGER_SQL_ALCHEMY_POOL_PRE_PING

        engine_args["pool_size"] = pool_size
        engine_args["pool_recycle"] = pool_recycle
        engine_args["pool_pre_ping"] = pool_pre_ping
        engine_args["max_overflow"] = max_overflow
        engine_args["poolclass"] = QueuePool
    else:
        engine_args = {"poolclass": NullPool}

    # Adding extra arguments
    if DB_LOGGER_SQL_ALCHEMY_CONNECTION_ARGS is not None:
        extra_args = DB_LOGGER_SQL_ALCHEMY_CONNECTION_ARGS.split(",")
        for arg in extra_args:
            if "=" not in arg:
                continue

            equal_idx = arg.index("=")
            key = arg[:equal_idx].strip()
            value = arg[equal_idx + 1 :].strip()  # noqa: E203
            if len(value) == 0 or len(key) == 0:
                continue
            engine_args[key] = value

    # Allow the user to specify an encoding for their DB otherwise default
    # to utf-8 so jobs & users with non-latin1 characters can still use
    # us.
    engine_args["encoding"] = DB_LOGGER_SQL_ENGINE_ENCODING
    # For Python2 we get back a newstr and need a str
    engine_args["encoding"] = engine_args["encoding"].__str__()

    return create_engine(DB_LOGGER_SQL_ALCHEMY_CONNECTION, **engine_args)


# The main db session. Used in all logging.
DB_LOGGER_ENGINE = create_db_logger_sqlalchemy_engine()

DBLoggerSession = scoped_session(
    sessionmaker(autocommit=False, autoflush=False, bind=DB_LOGGER_ENGINE, expire_on_commit=False)
)


def init_logger(reset=False):
    """Initialize the db logger database sqlalchemy models.

    Args:
        reset (bool, optional): If true then reset the tables. Defaults to False.
    """
    from airflow_db_logger.data import LoggerModelBase

    log.info(f"AirflowDBLogger is using: {DB_LOGGER_SQL_ALCHEMY_CONNECTION}")
    if reset:
        # NOTE: There is no promp for logs, when you reset, everything will reset always.
        log.info("Resetting db_logger tables...")
        LoggerModelBase.metadata.drop_all(DB_LOGGER_ENGINE)
    else:
        log.info("Initialzing db_logger tables...")
    LoggerModelBase.metadata.create_all(DB_LOGGER_ENGINE)

    log.info("AirflowDBLogger tables initialized.")


def check_cli_for_init_db():
    """Returns true if the cli command is for initializing the database."""
    # Old system
    if "initdb" in sys.argv or "upgradedb" in sys.argv or "resetdb" in sys.argv:
        init_logger("resetdb" in sys.argv)

    # New system
    if "db" in sys.argv and ("reset" in sys.argv or "init" in sys.argv or "upgrade" in sys.argv):
        init_logger("reset" in sys.argv)



import re
import os
import logging
from datetime import datetime
from weakref import WeakValueDictionary
from typing import Dict
from airflow_db_logger.handlers import DBLogStreamWriter, DBLogHandler, stderr_logger
from airflow_db_logger.exceptions import DBLoggerException
from airflow_db_logger.config import (
    DB_LOGGER_WRITE_TO_GCS_BUCKET,
    DB_LOGGER_WRITE_TO_GCS_PROJECT_ID,
    DB_LOGGER_WRITE_TO_GCS_MULTI_FILE_LOG,
)
from zthreading.decorators import collect_delayed_calls_async
from google.cloud.storage import Client

GOOGLE_STORAGE_RUI_REGEX = r"^(gs:\/\/|)([^\/]+)(\/(.*)|)$"


def to_storage_parts(gs_path):
    parts = re.findall(GOOGLE_STORAGE_RUI_REGEX, gs_path)
    bucket_name = parts[0][1]
    bucket_inner_path = parts[0][3] if len(parts[0]) > 2 else None
    return bucket_name, bucket_inner_path


class GCSFileLogProcessor:
    def __init__(self, filename: str) -> None:
        self.filename = filename
        self.pending_records = []
        self.bucket_name, self.bucket_inner_path = to_storage_parts(DB_LOGGER_WRITE_TO_GCS_BUCKET)

    def write(self, record):
        self.pending_records.append(record)
        self._write_async()

    def _compose_progressing_log_file_name(self, filename):
        filename, ext = os.path.splitext(filename)
        return f"{filename}.{datetime.now().strftime('%Y%m%d.%H%M%S.%f')}{ext}"

    @collect_delayed_calls_async(interval=0.1, on_error="write_async_error", use_daemon_thread=False)
    def _write_async(self):
        if len(self.pending_records) == 0:
            return
        try:
            client = Client(project=DB_LOGGER_WRITE_TO_GCS_PROJECT_ID)
            bucket_path = f"{self.bucket_inner_path}/{self.filename}"
            if DB_LOGGER_WRITE_TO_GCS_MULTI_FILE_LOG:
                bucket_path = self._compose_progressing_log_file_name(bucket_path)

            bucket = client.bucket(bucket_name=self.bucket_name)
            blob = bucket.blob(blob_name=bucket_path)

            records = self.pending_records
            self.pending_records = []

            if not DB_LOGGER_WRITE_TO_GCS_MULTI_FILE_LOG and blob.exists():
                current_log = blob.download_as_string().decode(encoding="utf-8").strip()
                if current_log:
                    records.insert(0, current_log)
                # Reset the blob
                blob = bucket.blob(blob_name=bucket_path)

            blob.upload_from_string("\n".join(records))

        except Exception as err:
            stderr_logger.error(
                f"Failed to flash to bucket @ {self.bucket_name}/{self.bucket_inner_path}/{self.filename}"
            )
            stderr_logger.error(err)

    def write_async_error(self, err: Exception):
        stderr_logger.error(err)


class GCSFileWriter(DBLogStreamWriter):
    def __init__(self, on_event=None) -> None:
        super().__init__(on_event=on_event)
        self._pending_loggers: Dict[str, GCSFileLogProcessor] = WeakValueDictionary()

    def get_file_collection_writer(self, filename: str):
        if filename not in self._pending_loggers:
            logger = GCSFileLogProcessor(filename)
            self._pending_loggers[filename] = logger
            return logger
        else:
            return self._pending_loggers[filename]

    def write(self, handler: DBLogHandler, record: logging.LogRecord):
        # Only applies when using context execution.
        if not handler.has_context:
            return

        filename = handler.get_logfile_subpath()
        assert isinstance(filename, str), DBLoggerException(
            f"Invalid filename when writing log @ {type(handler)}: {DB_LOGGER_WRITE_TO_GCS_BUCKET}/{filename}"
        )
        self.get_file_collection_writer(filename=filename).write(handler.format(record))



import os
import sys
import logging
from typing import Dict
from weakref import WeakValueDictionary
from zthreading.decorators import collect_delayed_calls_async
from airflow_db_logger.exceptions import DBLoggerException
from airflow_db_logger.config import BASE_LOG_FOLDER
from airflow_db_logger.handlers import DBLogStreamWriter, DBLogHandler, stderr_logger


class DBLogShellWriter(DBLogStreamWriter):
    def write(self, handler: "DBLogHandler", record: logging.LogRecord):
        msg = handler.format(record) + "\n"
        if record.levelno < logging.WARNING:
            sys.__stdout__.write(msg)
        else:
            sys.__stderr__.write(msg)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
__title__ = "airflow_db_logger"
__author__ = "Zav Shotan"

import sys
from copy import deepcopy
import airflow_db_logger.consts as consts

# THIS IS UGLY

LOGGING_CONFIG = consts.DB_LOGGER_LOGGING_CONFIG

from airflow.config_templates.airflow_local_settings import DEFAULT_LOGGING_CONFIG  # noqa
from airflow_db_logger.config import (  # noqa
    check_cli_for_init_db,  # noqa: E402
    DB_LOGGER_TASK_FORMATTER,  # noqa: E402
    DB_LOGGER_CONSOLE_FORMATTER,  # noqa: E402
    DB_LOGGER_WRITE_DAG_PROCESSING_TO_DB,  # noqa: E402
    DB_LOGGER_PROCESSER_LOG_LEVEL,
)


def update_config_from_defaults():
    consts.IS_LOADING_CONFIG

    if consts.IS_LOADING_CONFIG is True:
        return

    # Remove any other loads.
    consts.IS_LOADING_CONFIG = True

    processor_handler_config = {
        "class": "airflow_db_logger.handlers.StreamHandler",
        "formatter": DB_LOGGER_CONSOLE_FORMATTER,
        "level": DB_LOGGER_PROCESSER_LOG_LEVEL,
    }

    if DB_LOGGER_WRITE_DAG_PROCESSING_TO_DB:
        processor_handler_config = {
            "class": "airflow_db_logger.handlers.DBProcessLogHandler",
            "formatter": DB_LOGGER_CONSOLE_FORMATTER,
            "level": DB_LOGGER_PROCESSER_LOG_LEVEL,
        }

    LOGGING_CONFIG.update(deepcopy(DEFAULT_LOGGING_CONFIG))
    LOGGING_CONFIG["handlers"] = {
        "console": {
            "class": "airflow_db_logger.handlers.StreamHandler",
            "formatter": DB_LOGGER_CONSOLE_FORMATTER,
        },
        "task": {
            "class": "airflow_db_logger.handlers.DBTaskLogHandler",
            "formatter": DB_LOGGER_TASK_FORMATTER,
        },
        "processor": processor_handler_config,
    }

    # Checking for database initialization
    check_cli_for_init_db()


update_config_from_defaults()



import logging
import traceback
import os
import sys
from typing import Dict, List, Union
from zthreading.events import EventHandler, Event
from airflow.utils.helpers import parse_template_string
from airflow.models import TaskInstance
from sqlalchemy import asc, desc

from airflow_db_logger.exceptions import DBLoggerException
from airflow_db_logger.utils import get_calling_frame_objects_by_type
from airflow_db_logger.data import TaskExecutionLogRecord, DagFileProcessingLogRecord
from airflow_db_logger.config import (
    LOG_LEVEL,
    DBLoggerSession,
    DAGS_FOLDER,
    IS_RUNNING_DEBUG_EXECUTOR,
    DB_LOGGER_SHOW_REVERSE_ORDER,
    TASK_LOG_FILENAME_TEMPLATE,
    PROCESS_LOG_FILENAME_TEMPLATE,
    DB_LOGGER_WRITE_TO_FILES,
    DB_LOGGER_WRITE_TO_GCS_BUCKET,
    DB_LOGGER_WRITE_TO_SHELL,
)

stderr_logger = logging.getLogger(__file__)
stderr_logger.propagate = False
stderr_logger.handlers.clear()
stderr_handler = logging.StreamHandler(stream=sys.__stderr__)
stderr_handler.setFormatter(logging.Formatter(fmt="[%(asctime)s][DBLOGGER] [%(levelname)s] - %(message)s"))
stderr_logger.addHandler(stderr_handler)


class ExecutionLogTaskContextInfo:
    def __init__(self, task_instance: TaskInstance):
        super().__init__()
        self.dag_id = task_instance.dag_id
        self.task_id = task_instance.task_id
        self.execution_date = task_instance.execution_date
        self.try_number = task_instance.try_number


class DBLoggingEventHandler(EventHandler):
    log_event_name: str = "log"
    flush_event_name: str = "flush"
    close_event_name: str = "close"


class StreamHandler(logging.StreamHandler):
    def __init__(self, stream: str = None, level: str = "WARN") -> None:
        stream = stream or "stdout"
        self._use_stderr = "stderr" in stream
        logging.Handler.__init__(self, level=level)

    @property
    def stream(self):
        if self._use_stderr:
            return sys.__stderr__

        return sys.__stdout__


class DBLogStreamWriter(DBLoggingEventHandler):
    _stream_writers: List["DBLogStreamWriter"] = None

    def __init__(self, on_event=None) -> None:
        super().__init__(on_event=on_event)

    def emit_event(self, event: Event):
        """Override to handle logging events.

        Args:
            event (Event): The event object
        """
        if event.name == self.log_event_name:
            # assert isinstance(event.sender, DBLogHandler)
            self.write(handler=event.sender, record=event.args[0])

    def write(self, handler: "DBLogHandler", record: logging.LogRecord):
        """Abstract method. Writes a new log line.

        Args:
            handler (DBLogHandler): The log handler.
            record (any): The log record.

        Raises:
            NotImplementedError: [description]
        """
        raise NotImplementedError("This method is abstract, and should be overridden")

    @classmethod
    def _initialize_global_stream_writers(cls):
        cls._stream_writers = []
        if DB_LOGGER_WRITE_TO_FILES is not False:
            from airflow_db_logger.writers.local import DBLogFileWriter

            cls._stream_writers.append(DBLogFileWriter())
        if DB_LOGGER_WRITE_TO_GCS_BUCKET is not None:
            from airflow_db_logger.writers.gcs import GCSFileWriter

            cls._stream_writers.append(GCSFileWriter())

        if DB_LOGGER_WRITE_TO_SHELL is True:
            from airflow_db_logger.writers.shell import DBLogShellWriter

            cls._stream_writers.append(DBLogShellWriter())

    @classmethod
    def get_global_stream_writers(cls) -> List["DBLogStreamWriter"]:
        if cls._stream_writers is None:
            cls._initialize_global_stream_writers()
        return cls._stream_writers


class DBLogHandler(logging.Handler, DBLoggingEventHandler):
    def __init__(self, level: Union[str, int] = None):
        logging.Handler.__init__(self, level=level or LOG_LEVEL)
        EventHandler.__init__(self)
        self._db_session: DBLoggerSession = None
        for writer in DBLogStreamWriter.get_global_stream_writers():
            self.pipe(writer)

    @property
    def has_context(self):
        return self._db_session is not None

    @property
    def db_session(self) -> DBLoggerSession:
        return self._db_session

    def get_logfile_subpath(self):
        """Returns the file-path associated with the log handler. By default returns 'global.log',
        Override this method to provide handler specific filepath"""
        return "global.log"

    def set_context(self):
        """Initialize the db log configuration.

        Arguments:
            task_instance {task instance object} -- The task instace to write for.
        """
        if self._db_session is None:
            self._db_session = DBLoggerSession()

    def emit_handler_event(self, name: str, *args, **kwargs):
        """Emits an event. Any arguments sent after name, will
        be passed to the event action.

        Arguments:
            name {str} -- The name of the event to emit.
        """
        EventHandler.emit(self, name, *args, **kwargs)

    def emit(self, record: logging.LogRecord):
        """Emits a log record. Override this method to provide record handling.

        Arguments:
            record {any} -- The logging record.
        """
        self.emit_handler_event(self.log_event_name, record)

    def flush(self):
        """Waits for any unwritten logs to write to the db."""
        self.emit_handler_event(self.flush_event_name)
        if not self.has_context:
            return
        if self.db_session is not None:
            self.db_session.flush()

    def close(self):
        self.emit_handler_event(self.close_event_name)
        if not self.has_context:
            return
        """Stops and finalizes the log writing.
        """
        if self.db_session is not None:
            self.db_session.close()
            self._db_session = None


class DBTaskLogHandler(DBLogHandler):
    """
    DB Task log handler writes and reads task logs from the logging database
    (Defaults to the airflow database, unless otherwise defined)
    """

    subfolder_path: str = "tasks"

    def __init__(self, level: Union[str, int] = None):
        super().__init__(level=level)
        self.filename_template, self.filename_jinja_template = parse_template_string(TASK_LOG_FILENAME_TEMPLATE)
        self._task_context_info: ExecutionLogTaskContextInfo = None
        self._task_instance: TaskInstance = None
        self._logfile_subpath: str = None

    @property
    def task_context_info(self) -> ExecutionLogTaskContextInfo:
        assert self._task_context_info is not None, "Task instance was not defined while attempting to write task log"
        return self._task_context_info

    @property
    def has_task_context(self):
        return self._task_context_info is not None

    def get_logfile_subpath(self):
        """Returns the task log filename"""
        return self._logfile_subpath

    def _render_logfile_subpath(self):
        """Returns the task log sub filepath"""
        if self.filename_jinja_template:
            # render as jinja
            jinja_context = self._task_instance.get_template_context()
            jinja_context["ti"] = self.task_context_info
            jinja_context["try_number"] = self.task_context_info.try_number
            return self.filename_jinja_template.render(**jinja_context)
        else:
            # render direct
            return self.filename_template.format(
                dag_id=self.task_context_info.dag_id,
                task_id=self.task_context_info.task_id,
                execution_date=self.task_context_info.execution_date.isoformat(),
                try_number=self.task_context_info.try_number,
            )

    def set_context(self, task_instance):
        """Initialize the db log configuration.

        Arguments:
            task_instance {task instance object} -- The task instace to write for.
        """

        try:
            self._task_instance = task_instance
            self._task_context_info = ExecutionLogTaskContextInfo(task_instance)
            self._logfile_subpath = os.path.join(self.subfolder_path, self._render_logfile_subpath())
            super().set_context()
        except Exception as err:
            stderr_logger.error(err)

    def emit(self, record: logging.LogRecord):
        """Emits a log record.

        Arguments:
            record {any} -- The logging record.
        """

        # A fix to allow the debug executor to run also to the database.
        if IS_RUNNING_DEBUG_EXECUTOR and not self.has_task_context:
            ti: TaskInstance = get_calling_frame_objects_by_type(TaskInstance, first_only=True)
            if ti is not None:
                self.set_context(task_instance=ti)

        if self.has_task_context and self.has_context:
            try:
                db_record = TaskExecutionLogRecord(
                    self.task_context_info.dag_id,
                    self.task_context_info.task_id,
                    self.task_context_info.execution_date,
                    self.task_context_info.try_number,
                    self.format(record),
                )

                self.db_session.add(db_record)
                self.db_session.commit()
            except Exception:
                try:
                    self.db_session.rollback()
                except:
                    pass
                stderr_logger.error(traceback.format_exc())

        super().emit(record)

    def read(
        self,
        task_instance: TaskInstance,
        try_number: int = None,
        metadata: dict = None,
    ):
        """Read logs of given task instance from the database.

        Arguments:
            task_instance {TaskInstance} -- The task instance object

        Keyword Arguments:
            try_number {int} -- The run try number (default: {None})
            metadata {dict} -- Added metadata (default: {None})

        Raises:
            Exception: [description]

        Returns:
            List[str] -- A log array.
        """
        # Task instance increments its try number when it starts to run.
        # So the log for a particular task try will only show up when
        # try number gets incremented in DB, i.e logs produced the time
        # after cli run and before try_number + 1 in DB will not be displayed.
        db_session: DBLoggerSession = None
        try:
            db_session = DBLoggerSession()
            if try_number is None:
                next_try = task_instance.next_try_number
                try_numbers = list(range(1, next_try))
            elif try_number < 1:
                logs = [
                    "Error fetching the logs. Try number {} is invalid.".format(try_number),
                ]
                return logs
            else:
                try_numbers = [try_number]

            logs_by_try_number: Dict[int, List[TaskExecutionLogRecord]] = dict()

            log_records_query = (
                db_session.query(TaskExecutionLogRecord)
                .filter(TaskExecutionLogRecord.dag_id == task_instance.dag_id)
                .filter(TaskExecutionLogRecord.task_id == task_instance.task_id)
                .filter(TaskExecutionLogRecord.execution_date == task_instance.execution_date)
                .filter(TaskExecutionLogRecord.try_number.in_(try_numbers))
            )

            if DB_LOGGER_SHOW_REVERSE_ORDER is True:
                log_records_query = log_records_query.order_by(desc(TaskExecutionLogRecord.timestamp))
            else:
                log_records_query = log_records_query.order_by(asc(TaskExecutionLogRecord.timestamp))

            log_records = log_records_query.all()

            db_session.close()
            db_session = None

            log_record: TaskExecutionLogRecord = None

            # pull the records
            log_records: List[TaskExecutionLogRecord] = [r for r in log_records]

            for log_record in log_records:
                try_number = int(log_record.try_number)
                if try_number not in logs_by_try_number:
                    logs_by_try_number[try_number] = []
                logs_by_try_number[try_number].append(str(log_record.text))

            for try_number in logs_by_try_number.keys():
                logs_by_try_number[try_number] = "\n".join(logs_by_try_number[try_number])

            try_numbers.sort()
            logs = []
            metadata_array = []
            for try_number in try_numbers:
                logs.append(logs_by_try_number[try_number])
                metadata_array.append({"end_of_log": True})

            return logs, metadata_array

        except Exception:
            if db_session:
                try:
                    db_session.rollback()
                except Exception:
                    pass
            stderr_logger.error(traceback.format_exc())
            return [f"An error occurred while connecting to the database:\n" + f"{traceback.format_exc()}"], [
                {"end_of_log": True}
            ]
        finally:
            if db_session:
                db_session.close()


class DBProcessLogHandler(DBLogHandler):
    """
    FileProcessorHandler is a python log handler that handles
    dag processor logs. It creates and delegates log handling
    to a database
    """

    dags_subfolder_path: str = "dags"
    global_log_file: str = "global.log"
    subfolder_path: str = "process"

    def __init__(self, level: Union[str, int] = None):
        """
        :param filename_template: template filename string
        """

        super().__init__(level=level)
        self.dag_dir: str = os.path.expanduser(DAGS_FOLDER)
        self._log_filepath: str = os.path.join("process", "global.log")
        self.filename_template, self.filename_jinja_template = parse_template_string(PROCESS_LOG_FILENAME_TEMPLATE)

    @property
    def dag_relative_filepath(self) -> str:
        """Returns the relative path of the dag filename, to the dags directory."""
        return self._log_filepath

    def _render_relative_dag_filepath(self, filename: str):
        """Renders a dag log file relative to the dag filepath.

        Arguments:
            filename {str} -- The original filename

        Returns:
            str -- The display filename.
        """
        filename = os.path.relpath(filename, self.dag_dir)
        return f"{filename}.log"

    def set_context(self, filepath=None):
        """Initialize the dag log configuration.

        Arguments:
            dag_filepath {str} -- The path to the dag file.
        """

        try:
            self._log_filepath = (
                os.path.join(self.subfolder_path, self.global_log_file)
                if filepath is None
                else os.path.join(
                    self.subfolder_path, self.dags_subfolder_path, self._render_relative_dag_filepath(filepath)
                )
            )
            self._db_session = DBLoggerSession()
        except Exception:
            logging.error("Failed to initialize process logger contexts")
            traceback.print_exc()
            # logging.error(err)

    def get_logfile_subpath(self):
        assert self._log_filepath, DBLoggerException(
            "Cannot write dag logs to file, dag_relative_filepath is empty. Context not set."
        )
        return self._log_filepath

    def emit(self, record: logging.LogRecord):
        """Emits a log record.

        Arguments:
            record {any} -- The logging record.
        """
        if not self.has_context:
            self.set_context()

        db_record_message = self.format(record)
        try:
            db_record = DagFileProcessingLogRecord(self._log_filepath, db_record_message)
            self.db_session.add(db_record)
            self.db_session.commit()
        except Exception:
            try:
                self.db_session.rollback()
            except:
                pass
            stderr_logger.error(f"Error while attempting to log ({self._log_filepath}): {db_record_message}")
            stderr_logger.error(traceback.format_exc())

        super().emit(record)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
from setuptools import setup

here = os.path.abspath(os.path.dirname(__file__))


def get_version():
    version_file_path = os.path.join(here, "package_version.txt")
    if not os.path.isfile(version_file_path):
        return "debug"
    version = None
    with open(version_file_path, "r") as raw:
        version = raw.read()

    return version


setup(
    name="airflow_db_logger",
    version=get_version(),
    description="An airflow logger that stores its results in a database given an SQLAlchemy connection.",
    long_description="Please see readme.md @ https://github.com/LamaAni/AirflowDBLogger",
    classifiers=[],
    author="Zav Shotan",
    author_email="",
    url="https://github.com/LamaAni/AirflowDBLogger",
    packages=["airflow_db_logger", "airflow_db_logger/writers"],
    platforms="any",
    license="docs/LICENSE",
    install_requires=[
        "sqlalchemy>=0.23.1",
        "zthreading>=0.1.15",
    ],
    python_requires=">=3.6",
    include_package_data=True,
)
